#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Isoelastics management"""
from __future__ import division, unicode_literals

import warnings

import pathlib
from pkg_resources import resource_filename

import numpy as np

from .. import definitions as dfn
from ..features import emodulus as feat_emod
from ..features import emodulus_viscosity as feat_emod_visc

ISOFILES = ["isoel-analytical-area_um-deform.txt",
            "isoel-numerical-area_um-deform.txt",
            ]
ISOFILES = [resource_filename("dclab.isoelastics", _if) for _if in ISOFILES]

VALID_METHODS = ["analytical", "numerical"]


class IsoelasticsEmodulusMeaninglessWarning(UserWarning):
    pass


class Isoelastics(object):
    def __init__(self, paths=[]):

        self._data = IsoelasticsDict()

        for path in paths:
            self.load_data(path)

    def _add(self, isoel, col1, col2, method, meta):
        """Convenience method for population self._data"""
        self._data[method][col1][col2]["isoelastics"] = isoel
        self._data[method][col1][col2]["meta"] = meta

        # Use advanced slicing to flip the data columns
        isoel_flip = [iso[:, [1, 0, 2]] for iso in isoel]
        self._data[method][col2][col1]["isoelastics"] = isoel_flip
        self._data[method][col2][col1]["meta"] = meta

    def add(self, isoel, col1, col2, channel_width,
            flow_rate, viscosity, method):
        """Add isoelastics

        Parameters
        ----------
        isoel: list of ndarrays
            Each list item resembles one isoelastic line stored
            as an array of shape (N,3). The last column contains
            the emodulus data.
        col1: str
            Name of the first feature of all isoelastics
            (e.g. isoel[0][:,0])
        col2: str
            Name of the second feature of all isoelastics
            (e.g. isoel[0][:,1])
        channel_width: float
            Channel width in µm
        flow_rate: float
            Flow rate through the channel in µl/s
        viscosity: float
            Viscosity of the medium in mPa*s
        method: str
            The method used to compute the isoelastics
            (must be one of `VALID_METHODS`).

        Notes
        -----
        The following isoelastics are automatically added for
        user convenience:
        - isoelastics with `col1` and `col2` interchanged
        - isoelastics for circularity if deformation was given
        """
        if method not in VALID_METHODS:
            validstr = ",".join(VALID_METHODS)
            raise ValueError("`method` must be one of {}!".format(validstr))
        for col in [col1, col2]:
            if col not in dfn.scalar_feature_names:
                raise ValueError("Not a valid feature name: {}".format(col))

        meta = [channel_width, flow_rate, viscosity]

        # Add the feature data
        self._add(isoel, col1, col2, method, meta)

        # Also add the feature data for circularity
        if "deform" in [col1, col2]:
            col1c, col2c = col1, col2
            if col1c == "deform":
                deform_ax = 0
                col1c = "circ"
            else:
                deform_ax = 1
                col2c = "circ"
            iso_circ = []
            for iso in isoel:
                iso = iso.copy()
                iso[:, deform_ax] = 1 - iso[:, deform_ax]
                iso_circ.append(iso)
            self._add(iso_circ, col1c, col2c, method, meta)

    @staticmethod
    def add_px_err(isoel, col1, col2, px_um, inplace=False):
        """Undo pixelation correction

        Isoelasticity lines are already corrected for pixelation
        effects as described in

        Mapping of Deformation to Apparent Young's Modulus
        in Real-Time Deformability Cytometry
        Christoph Herold, arXiv:1704.00572 [cond-mat.soft] (2017)
        https://arxiv.org/abs/1704.00572.

        If the isoealsticity lines are displayed with deformation data
        that are not corrected, then the lines must be "un"-corrected,
        i.e. the pixelation error must be added to the lines to match
        the experimental data.

        Parameters
        ----------
        isoel: list of 2d ndarrays of shape (N, 3)
            Each item in the list corresponds to one isoelasticity
            line. The first column is defined by `col1`, the second
            by `col2`, and the third column is the emodulus.
        col1, col2: str
            Define the fist to columns of each isoelasticity line.
            One of ["area_um", "circ", "deform"]
        px_um: float
            Pixel size [µm]
        """
        Isoelastics.check_col12(col1, col2)
        if "deform" in [col1, col2]:
            # add error for deformation
            sign = +1
        else:
            # subtract error for circularity
            sign = -1
        if col1 == "area_um":
            area_ax = 0
            deci_ax = 1
        else:
            area_ax = 1
            deci_ax = 0

        new_isoel = []
        for iso in isoel:
            iso = np.array(iso, copy=not inplace)
            ddeci = feat_emod.corrpix_deform_delta(area_um=iso[:, area_ax],
                                                   px_um=px_um)
            iso[:, deci_ax] += sign * ddeci
            new_isoel.append(iso)
        return new_isoel

    @staticmethod
    def check_col12(col1, col2):
        if (col1 not in ["area_um", "circ", "deform"] or
                col2 not in ["area_um", "circ", "deform"]):
            raise ValueError("Columns must be one of: area_um, circ, deform!")
        if col1 == col2:
            raise ValueError("Columns are the same!")
        if "area_um" not in [col1, col2]:
            # avoid [circ, deform]
            raise ValueError("One column must be set to 'area_um'!")

    @staticmethod
    def convert(isoel, col1, col2,
                channel_width_in, channel_width_out,
                flow_rate_in, flow_rate_out,
                viscosity_in, viscosity_out,
                inplace=False):
        """Convert isoelastics in area_um-deform space

        Parameters
        ----------
        isoel: list of 2d ndarrays of shape (N, 3)
            Each item in the list corresponds to one isoelasticity
            line. The first column is defined by `col1`, the second
            by `col2`, and the third column is the emodulus.
        col1, col2: str
            Define the fist to columns of each isoelasticity line.
            One of ["area_um", "circ", "deform"]
        channel_width_in: float
            Original channel width [µm]
        channel_width_out: float
            Target channel width [µm]
        flow_rate_in: float
            Original flow rate [µl/s]
        flow_rate_in: float
            Target flow rate [µl/s]
        viscosity_in: float
            Original viscosity [mPa*s]
        viscosity_out: float
            Target viscosity [mPa*s]

        Notes
        -----
        If only the positions of the isoelastics are of interest and
        not the value of the elastic modulus, then it is sufficient
        to supply values for the channel width and set the values
        for flow rate and viscosity to a constant (e.g. 1).

        See Also
        --------
        dclab.features.emodulus.convert: conversion method used
        """
        Isoelastics.check_col12(col1, col2)
        if col1 == "area_um":
            area_ax = 0
            defo_ax = 1
        else:
            area_ax = 1
            defo_ax = 0

        new_isoel = []

        for iso in isoel:
            iso = np.array(iso, copy=not inplace)
            feat_emod.convert(area_um=iso[:, area_ax],
                              deform=iso[:, defo_ax],
                              emodulus=iso[:, 2],
                              channel_width_in=channel_width_in,
                              channel_width_out=channel_width_out,
                              flow_rate_in=flow_rate_in,
                              flow_rate_out=flow_rate_out,
                              viscosity_in=viscosity_in,
                              viscosity_out=viscosity_out,
                              inplace=True)
            new_isoel.append(iso)
        return new_isoel

    def get(self, col1, col2, method, channel_width, flow_rate=None,
            viscosity=None, add_px_err=False, px_um=None):
        """Get isoelastics

        Parameters
        ----------
        col1: str
            Name of the first feature of all isoelastics
            (e.g. isoel[0][:,0])
        col2: str
            Name of the second feature of all isoelastics
            (e.g. isoel[0][:,1])
        method: str
            The method used to compute the isoelastics
            (must be one of `VALID_METHODS`).
        channel_width: float
            Channel width in µm
        flow_rate: float or `None`
            Flow rate through the channel in µl/s. If set to
            `None`, the flow rate of the imported data will
            be used (only do this if you do not need the
            correct values for elastic moduli).
        viscosity: float or `None`
            Viscosity of the medium in mPa*s. If set to
            `None`, the flow rate of the imported data will
            be used (only do this if you do not need the
            correct values for elastic moduli).
        add_px_err: bool
            If True, add pixelation errors according to
            C. Herold (2017), https://arxiv.org/abs/1704.00572
        px_um: float
            Pixel size [µm], used for pixelation error computation

        See Also
        --------
        dclab.features.emodulus.convert: conversion in-between
            channel sizes and viscosities
        dclab.features.emodulus.corrpix_deform_delta: pixelation
            error that is applied to the deformation data
        """
        if method not in VALID_METHODS:
            validstr = ",".join(VALID_METHODS)
            raise ValueError("`method` must be one of {}!".format(validstr))
        for col in [col1, col2]:
            if col not in dfn.scalar_feature_names:
                raise ValueError("Not a valid feature name: {}".format(col))

        if "isoelastics" not in self._data[method][col2][col1]:
            msg = "No isoelastics matching {}, {}, {}".format(col1, col2,
                                                              method)
            raise KeyError(msg)

        isoel = self._data[method][col1][col2]["isoelastics"]
        meta = self._data[method][col1][col2]["meta"]

        if flow_rate is None:
            flow_rate = meta[1]

        if viscosity is None:
            viscosity = meta[2]

        isoel_ret = self.convert(isoel, col1, col2,
                                 channel_width_in=meta[0],
                                 channel_width_out=channel_width,
                                 flow_rate_in=meta[1],
                                 flow_rate_out=flow_rate,
                                 viscosity_in=meta[2],
                                 viscosity_out=viscosity,
                                 inplace=False)

        if add_px_err:
            self.add_px_err(isoel=isoel_ret,
                            col1=col1,
                            col2=col2,
                            px_um=px_um,
                            inplace=True)

        return isoel_ret

    def get_with_rtdcbase(self, col1, col2, method, dataset,
                          viscosity=None, add_px_err=False):
        """Convenience method that extracts the metadata from RTDCBase

        Parameters
        ----------
        col1: str
            Name of the first feature of all isoelastics
            (e.g. isoel[0][:,0])
        col2: str
            Name of the second feature of all isoelastics
            (e.g. isoel[0][:,1])
        method: str
            The method used to compute the isoelastics
            (must be one of `VALID_METHODS`).
        dataset: dclab.rtdc_dataset.RTDCBase
            The dataset from which to obtain the metadata.
        viscosity: float, `None`, or False
            Viscosity of the medium in mPa*s. If set to
            `None`, the viscosity is computed from the meta
            data (medium, flow rate, channel width, temperature)
            in the [setup] config section. If this is not possible,
            the flow rate of the imported data is used and a warning
            will be issued.
        add_px_err: bool
            If True, add pixelation errors according to
            C. Herold (2017), https://arxiv.org/abs/1704.00572
        """
        cfg = dataset.config
        if viscosity is None:
            if "temperature" in cfg["setup"] and "medium" in cfg["setup"]:
                viscosity = feat_emod_visc.get_viscosity(
                    medium=cfg["setup"]["medium"],
                    channel_width=cfg["setup"]["channel width"],
                    flow_rate=cfg["setup"]["flow rate"],
                    temperature=cfg["setup"]["temperature"])
            else:
                warnings.warn("Computing emodulus data for isoelastics from "
                              + "RTDCBase is not possible. Isoelastics will "
                              + "not have correct emodulus values (this is "
                              + "not relevant for plotting).",
                              IsoelasticsEmodulusMeaninglessWarning)
        return self.get(col1=col1,
                        col2=col2,
                        method=method,
                        channel_width=cfg["setup"]["channel width"],
                        flow_rate=cfg["setup"]["flow rate"],
                        viscosity=viscosity,
                        add_px_err=add_px_err,
                        px_um=cfg["imaging"]["pixel size"])

    def load_data(self, path):
        """Load isoelastics from a text file

        The text file is loaded with `numpy.loadtxt` and must have
        three columns, representing the two data columns and the
        elastic modulus with units defined in `definitions.py`.
        The file header must have a section defining meta data of the
        content like so:

            # [...]
            #
            # - column 1: area_um
            # - column 2: deform
            # - column 3: emodulus
            # - channel width [um]: 20
            # - flow rate [ul/s]: 0.04
            # - viscosity [mPa*s]: 15
            # - method: analytical
            #
            # [...]


        Parameters
        ----------
        path: str
            Path to a isoelastics text file
        """
        path = pathlib.Path(path).resolve()
        # Get metadata
        meta = {}
        with path.open("r", errors='replace') as fd:
            while True:
                line = fd.readline().strip()
                if line.startswith("# - "):
                    line = line.strip("#- ")
                    var, val = line.split(":")
                    if val.strip().replace(".", "").isdigit():
                        # channel width, flow rate, viscosity
                        val = float(val)
                    else:
                        # columns, calculation
                        val = val.strip().lower()
                    meta[var.strip()] = val
                elif line and not line.startswith("#"):
                    break

        assert meta["column 1"] in dfn.scalar_feature_names
        assert meta["column 2"] in dfn.scalar_feature_names
        assert meta["column 3"] == "emodulus"
        assert meta["method"] in VALID_METHODS

        # Load isoelasics
        with path.open("rb") as isfd:
            isodata = np.loadtxt(isfd)

        # Slice out individual isoelastics
        emoduli = np.unique(isodata[:, 2])
        isoel = []
        for emod in emoduli:
            where = isodata[:, 2] == emod
            isoel.append(isodata[where])

        # Add isoelastics to instance
        self.add(isoel=isoel,
                 col1=meta["column 1"],
                 col2=meta["column 2"],
                 channel_width=meta["channel width [um]"],
                 flow_rate=meta["flow rate [ul/s]"],
                 viscosity=meta["viscosity [mPa*s]"],
                 method=meta["method"])


class IsoelasticsDict(dict):
    def __getitem__(self, key):
        if key in VALID_METHODS + dfn.scalar_feature_names:
            if key not in self:
                self[key] = IsoelasticsDict()
        return super(IsoelasticsDict, self).__getitem__(key)


def get_default():
    """Return default isoelasticity lines"""
    return Isoelastics(ISOFILES)



# Keys are the feature names dclab and
# values are the feature names in the tdms file format.
dclab2tdms = {
    "area_cvx": "area",
    "area_msd": "raw area",
    "bright_avg": "Brightness",
    "bright_sd": "BrightnessSD",
    "circ": "circularity",
    "fl1_area": "FL1area",
    "fl1_dist": "FL1dpeaks",
    "fl1_max": "FL1max",
    "fl1_npeaks": "FL1npeaks",
    "fl1_pos": "FL1pos",
    "fl1_width": "FL1width",
    "fl2_area": "FL2area",
    "fl2_dist": "FL2dpeaks",
    "fl2_max": "FL2max",
    "fl2_npeaks": "FL2npeaks",
    "fl2_pos": "FL2pos",
    "fl2_width": "FL2width",
    "fl3_area": "FL3area",
    "fl3_dist": "FL3dpeaks",
    "fl3_max": "FL3max",
    "fl3_npeaks": "FL3npeaks",
    "fl3_pos": "FL3pos",
    "fl3_width": "FL3width",
    "g_force": "gValue",
    "frame": "time",  # [sic]
    "inert_ratio_cvx": "InertiaRatio",
    "inert_ratio_raw": "InertiaRatioRaw",
    "nevents": "NrOfCells",
    "pc1": "PC1",
    "pc2": "PC2",
    "pos_x": "x",
    "pos_y": "y",
    "size_x": "ax2",
    "size_y": "ax1",
    "temp": "temp",
    "temp_amb": "temp_amb",
}

# Add lower-case userdef features
for _i in range(10):
    dclab2tdms["userdef{}".format(_i)] = "userDef{}".format(_i)

# inverse of dclab2tdms
tdms2dclab = {}
for kk in dclab2tdms:
    tdms2dclab[dclab2tdms[kk]] = kk

# Add capitalized userdef features as well.
# see https://github.com/ZELLMECHANIK-DRESDEN/ShapeOut/issues/212
for _i in range(10):
    tdms2dclab["UserDef{}".format(_i)] = "userdef{}".format(_i)

# traces_tdms file definitions
# The second feature should not contain duplicates! - even if the
# entries in the first features are different.
tr_data_map = {
    "fl1_raw": ["fluorescence traces", "FL1raw"],
    "fl2_raw": ["fluorescence traces", "FL2raw"],
    "fl3_raw": ["fluorescence traces", "FL3raw"],
    "fl1_median": ["fluorescence traces", "FL1med"],
    "fl2_median": ["fluorescence traces", "FL2med"],
    "fl3_median": ["fluorescence traces", "FL3med"],
}

#: for param.ini and camera.ini
configmap = {
    "experiment": {
        "run index": ("General", "Measurement Number"),
        "sample": ("General", "Sample Name"),
        "date": ("General", "Date [YYYY-MM-DD]"),
        "time": ("General", "Start Time [hh:mm:ss]"),
    },
    # All special keywords related to RT-FDC
    "fluorescence": {
        "bit depth": ("FLUOR", "Bitdepthraw"),
        "channel count": ("FLUOR", "FL-Channels"),
        "channel 1 name": ("FLUOR", "Channel 1 Name"),
        "channel 2 name": ("FLUOR", "Channel 2 Name"),
        "channel 3 name": ("FLUOR", "Channel 3 Name"),
        "channels installed": ("FLUOR", "Channels Installed"),
        "laser 1 power": ("FLUOR", "Laser1 488 nm Power [%]"),
        "laser 2 power": ("FLUOR", "Laser2 561 nm Power [%]"),
        "laser 3 power": ("FLUOR", "Laser3 640 nm Power [%]"),
        "laser count": ("FLUOR", "Laser Count"),
        "lasers installed": ("FLUOR", "Lasers Installed"),
        "sample rate": ("FLUOR", "Samplerate [sps]"),
        "samples per event": ("FLUOR", "Samples Per Event"),
        "signal max": ("FLUOR", "ADCmax [V]"),
        "signal min": ("FLUOR", "ADCmin [V]"),
        "trace median": ("FLUOR", "Trace Median"),
    },
    # All tdms-related parameters
    "fmt_tdms": {
        "video frame offset": ("General", "video frame offset"),
    },
    # All imaging-related keywords
    "imaging": {
        "flash device": "LED (undefined)",
        "flash duration": ("General", "Shutter Time LED [us]"),
        "frame rate": ("Framerate", "Frame Rate"),
        "pixel size": ("Image", "Pix Size"),
        "roi position x": ("ROI", "x-pos"),
        "roi position y": ("ROI", "y-pos"),
        "roi size x": ("ROI", "width"),
        "roi size y": ("ROI", "height"),
    },
    # All parameters for online contour extraction from the event images
    "online_contour": {
        "bin area min": ("Image", "Trig Thresh"),
        "bin kernel": ("Image", "Bin Ops"),
        "bin threshold": ("Image", "Thresh"),
        "image blur": ("Image", "Blur"),
        "no absdiff": ("Image", "Diff_Method"),
    },
    # All online filters
    "online_filter": {
        "aspect min": ("Image", "Cell Aspect Min"),
        "aspect max": ("Image", "Cell Aspect Max"),
        "size_x max": ("Image", "Cell Max Length"),
        "size_y max": ("Image", "Cell Max Height"),
    },
    # All setup-related keywords, except imaging
    "setup": {
        "channel width": ("General", "Channel width [um]"),
        "chip region": ("General", "Region"),
        "flow rate": ("General", "Flow Rate [ul/s]"),
        "flow rate sample": ("General", "Sample Flow Rate [ul/s]"),
        "flow rate sheath": ("General", "Sheath Flow Rate [ul/s]"),
        "identifier": ("General", "Identifier"),
        "medium": ("General", "Buffer Medium"),
        "module composition": ("Image", "Setup"),
        "software version": ("General", "Software Version"),
    },
}

#: for SoftwareSettings.ini
config_map_set = {
    "online_contour": {
        "bin kernel": ("MaintenanceD", "bin_ops"),
        "image blur": ("MaintenanceD", "blur"),
    },
    "fluorescence": {
        "laser 1 power": ("Fluorescence", "Laser1_Power"),
        "laser 2 power": ("Fluorescence", "Laser2_Power"),
        "laser 3 power": ("Fluorescence", "Laser3_Power"),
        "trace median": ("Maintenance", "median_pmt"),
    },
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import io
import os
from os.path import join
import shutil
import tempfile

import numpy as np
import pytest

import dclab
from dclab import dfn, new_dataset

from helper_methods import example_data_dict, retrieve_data, \
    example_data_sets, cleanup


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.export.NoImageWarning')
def test_avi_export():
    ds = new_dataset(retrieve_data(example_data_sets[1]))
    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.avi")
    ds.export.avi(path=f1)
    assert os.stat(
        f1)[6] > 1e4, "Resulting file to small, Something went wrong!"
    shutil.rmtree(edest, ignore_errors=True)
    cleanup()


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.export.NoImageWarning')
def test_avi_override():
    ds = new_dataset(retrieve_data(example_data_sets[1]))

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.avi")
    ds.export.avi(f1, override=True)
    try:
        ds.export.avi(f1[:-4], override=False)
    except OSError:
        pass
    else:
        raise ValueError("Should append .avi and not override!")

    # cleanup
    shutil.rmtree(edest, ignore_errors=True)
    cleanup()


def test_avi_no_images():
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=127, keys=keys)
    ds = dclab.new_dataset(ddict)

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.avi")
    try:
        ds.export.avi(f1)
    except OSError:
        pass
    else:
        raise ValueError("There should be no image data to write!")
    shutil.rmtree(edest, ignore_errors=True)


def test_fcs_export():
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=222, keys=keys)
    ds = dclab.new_dataset(ddict)

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.fcs")
    f2 = join(edest, "test_unicode.fcs")

    ds.export.fcs(f1, keys, override=True)
    ds.export.fcs(f2, [u"area_um", u"deform", u"time",
                       u"frame", u"fl3_width"], override=True)

    with io.open(f1, mode="rb") as fd:
        a1 = fd.read()

    with io.open(f2, mode="rb") as fd:
        a2 = fd.read()

    assert a1 == a2
    assert len(a1) != 0

    # cleanup
    shutil.rmtree(edest, ignore_errors=True)


def test_fcs_override():
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=212, keys=keys)
    ds = dclab.new_dataset(ddict)

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.fcs")
    ds.export.fcs(f1, keys, override=True)
    try:
        ds.export.fcs(f1[:-4], keys, override=False)
    except OSError:
        pass
    else:
        raise ValueError("Should append .fcs and not override!")

    # cleanup
    shutil.rmtree(edest, ignore_errors=True)


def test_fcs_not_filtered():
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=127, keys=keys)
    ds = dclab.new_dataset(ddict)

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.tsv")
    ds.export.fcs(f1, keys, filtered=False)

    # cleanup
    shutil.rmtree(edest, ignore_errors=True)


def test_hdf5():
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=127, keys=keys)
    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5.rtdc")
    ds1.export.hdf5(f1, keys)

    ds2 = dclab.new_dataset(f1)
    assert ds1 != ds2
    assert np.allclose(ds2["area_um"], ds1["area_um"])
    assert np.allclose(ds2["deform"], ds1["deform"])
    assert np.allclose(ds2["time"], ds1["time"])
    assert np.allclose(ds2["frame"], ds1["frame"])
    assert np.allclose(ds2["fl3_width"], ds1["fl3_width"])

    # cleanup
    shutil.rmtree(edest, ignore_errors=True)


def test_hdf5_contour_image_trace():
    N = 65
    keys = ["contour", "image", "trace"]
    ddict = example_data_dict(size=N, keys=keys)

    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5_image.rtdc")
    ds1.export.hdf5(f1, keys, filtered=False)
    ds2 = dclab.new_dataset(f1)

    for ii in range(N):
        assert np.all(ds1["image"][ii] == ds2["image"][ii])
        assert np.all(ds1["contour"][ii] == ds2["contour"][ii])

    for key in dfn.FLUOR_TRACES:
        assert np.all(ds1["trace"][key] == ds2["trace"][key])

    # cleanup
    shutil.rmtree(edest, ignore_errors=True)


def test_hdf5_filtered():
    N = 10
    keys = ["area_um", "image"]
    ddict = example_data_dict(size=N, keys=keys)
    ddict["image"][3] = np.arange(10 * 20, dtype=np.uint8).reshape(10, 20) + 22

    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.filter.manual[2] = False
    ds1.apply_filter()
    fta = ds1.filter.manual.copy()

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5_filtered.rtdc")
    ds1.export.hdf5(f1, keys)

    ds2 = dclab.new_dataset(f1)

    assert ds1 != ds2
    assert np.allclose(ds2["area_um"], ds1["area_um"][fta])
    assert np.allclose(ds2["image"][2], ds1["image"][3])
    assert np.all(ds2["image"][2] != ds1["image"][2])

    # cleanup
    shutil.rmtree(edest, ignore_errors=True)


def test_hdf5_filtered_index():
    """Make sure that exported index is always re-enumerated"""
    N = 10
    keys = ["area_um", "deform", "index"]
    ddict = example_data_dict(size=N, keys=keys)
    ddict["index"] = np.arange(1, N+1)

    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.filter.manual[2] = False
    ds1.apply_filter()

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5_filtered.rtdc")
    ds1.export.hdf5(f1, keys)

    ds2 = dclab.new_dataset(f1)

    assert len(ds2) == N - 1
    assert np.all(ds2["index"] == np.arange(1, N))
    assert ds2.config["experiment"]["event count"] == N - 1

    # cleanup
    shutil.rmtree(edest, ignore_errors=True)


def test_hdf5_override():
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=212, keys=keys)
    ds = dclab.new_dataset(ddict)

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.rtdc")
    ds.export.hdf5(f1, keys, override=True)
    try:
        ds.export.hdf5(f1[:-5], keys, override=False)
    except OSError:
        pass
    else:
        raise ValueError("Should append .rtdc and not override!")

    # cleanup
    shutil.rmtree(edest, ignore_errors=True)


def test_tsv_export():
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=222, keys=keys)
    ds = dclab.new_dataset(ddict)

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.tsv")
    f2 = join(edest, "test_unicode.tsv")

    ds.export.tsv(f1, keys, override=True)
    ds.export.tsv(f2, [u"area_um", u"deform", u"time",
                       u"frame", u"fl3_width"], override=True)

    with io.open(f1) as fd:
        a1 = fd.read()

    with io.open(f2) as fd:
        a2 = fd.read()

    assert a1 == a2
    assert len(a1) != 0

    # cleanup
    shutil.rmtree(edest, ignore_errors=True)


def test_tsv_override():
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=212, keys=keys)
    ds = dclab.new_dataset(ddict)

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.tsv")
    ds.export.tsv(f1, keys, override=True)
    try:
        ds.export.tsv(f1[:-4], keys, override=False)
    except OSError:
        pass
    else:
        raise ValueError("Should append .tsv and not override!")

    # cleanup
    shutil.rmtree(edest, ignore_errors=True)


def test_tsv_not_filtered():
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=127, keys=keys)
    ds = dclab.new_dataset(ddict)

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.tsv")
    ds.export.tsv(f1, keys, filtered=False)

    # cleanup
    shutil.rmtree(edest, ignore_errors=True)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Content-based downsampling of ndarrays"""
from __future__ import division, print_function, unicode_literals

import numpy as np

from .cached import Cache


def downsample_rand(a, samples, remove_invalid=False, ret_idx=False):
    """Downsampling by randomly removing points

    Parameters
    ----------
    a: 1d ndarray
        The input array to downsample
    samples: int
        The desired number of samples
    remove_invalid: bool
        Remove nan and inf values before downsampling
    ret_idx: bool
        Also return a boolean array that corresponds to the
        downsampled indices in `a`.

    Returns
    -------
    dsa: 1d ndarray of size `samples`
        The pseudo-randomly downsampled array `a`
    idx: 1d boolean array with same shape as `a`
        Only returned if `ret_idx` is True.
        A boolean array such that `a[idx] == dsa`
    """
    # fixed random state for this method
    rs = np.random.RandomState(seed=47).get_state()
    np.random.set_state(rs)

    samples = int(samples)

    if remove_invalid:
        # slice out nans and infs
        bad = np.isnan(a) | np.isinf(a)
        pool = a[~bad]
    else:
        pool = a

    if samples and (samples < pool.shape[0]):
        keep = np.zeros_like(pool, dtype=bool)
        keep_ids = np.random.choice(np.arange(pool.size),
                                    size=samples,
                                    replace=False)
        keep[keep_ids] = True
        dsa = pool[keep]
    else:
        keep = np.ones_like(pool, dtype=bool)
        dsa = pool

    if remove_invalid:
        # translate the kept values back to the original array
        idx = np.zeros(a.size, dtype=bool)
        idx[~bad] = keep
    else:
        idx = keep

    if ret_idx:
        return dsa, idx
    else:
        return dsa


@Cache
def downsample_grid(a, b, samples, remove_invalid=False, ret_idx=False):
    """Content-based downsampling for faster visualization

    The arrays `a` and `b` make up a 2D scatter plot with high
    and low density values. This method takes out points at
    indices with high density.

    Parameters
    ----------
    a, b: 1d ndarrays
        The input arrays to downsample
    samples: int
        The desired number of samples
    remove_invalid: bool
        Remove nan and inf values before downsampling; if set to
        `True`, the actual number of samples returned might be
        smaller than `samples` due to infinite or nan values.
    ret_idx: bool
        Also return a boolean array that corresponds to the
        downsampled indices in `a` and `b`.

    Returns
    -------
    dsa, dsb: 1d ndarrays of shape (samples,)
        The arrays `a` and `b` downsampled by evenly selecting
        points and pseudo-randomly adding or removing points
        to match `samples`.
    idx: 1d boolean array with same shape as `a`
        Only returned if `ret_idx` is True.
        A boolean array such that `a[idx] == dsa`
    """
    # fixed random state for this method
    rs = np.random.RandomState(seed=47).get_state()

    if remove_invalid:
        # Remove nan and inf values straight from the beginning.
        # This might results in arrays smaller than `samples`,
        # but it makes sure that no inf/nan values will be plotted.
        bad = np.isnan(a) | np.isinf(a) | np.isnan(b) | np.isinf(b)
        ad = a[~bad]
        bd = b[~bad]
    else:
        bad = np.zeros_like(a, dtype=bool)
        ad = a
        bd = b

    keep = np.ones_like(a, dtype=bool)
    keep[bad] = False

    samples = int(samples)

    if samples and samples < ad.size:
        # The events to keep
        keepd = np.zeros_like(ad, dtype=bool)

        # 1. Produce evenly distributed samples
        # Choosing grid-size:
        # - large numbers tend to show actual structures of the sample,
        #   which is not desired for plotting
        # - small numbers tend will not result in too few samples and,
        #   in order to reach the desired samples, the data must be
        #   upsampled again.
        # 300 is about the size of the plot in marker sizes and yields
        # good results.
        grid_size = 300
        xpx = norm(ad, ad, bd) * grid_size
        ypx = norm(bd, bd, ad) * grid_size
        # The events on the grid to process
        toproc = np.ones((grid_size, grid_size), dtype=bool)

        for ii in range(xpx.size):
            xi = xpx[ii]
            yi = ypx[ii]
            # filter for overlapping events
            # Note that `valid` is used here to promote only valid
            # events in this step. However, in step 2, invalid events
            # could be added back. To avoid this scenario, the
            # parameter `remove_invalid` should be set to True.
            if valid(xi, yi) and toproc[int(xi-1), int(yi-1)]:
                toproc[int(xi-1), int(yi-1)] = False
                # include event
                keepd[ii] = True

        # 2. Make sure that we reach `samples` by adding or
        # removing events.
        diff = np.sum(keepd) - samples
        if diff > 0:
            # Too many samples
            rem_indices = np.where(keepd)[0]
            np.random.set_state(rs)
            rem = np.random.choice(rem_indices,
                                   size=diff,
                                   replace=False)
            keepd[rem] = False
        elif diff < 0:
            # Not enough samples
            add_indices = np.where(~keepd)[0]
            np.random.set_state(rs)
            add = np.random.choice(add_indices,
                                   size=abs(diff),
                                   replace=False)
            keepd[add] = True

        assert np.sum(keepd) == samples, "sanity check"
        asd = ad[keepd]
        bsd = bd[keepd]
        assert np.allclose(ad[keepd], asd, equal_nan=True), "sanity check"
        assert np.allclose(bd[keepd], bsd, equal_nan=True), "sanity check"
        keep[~bad] = keepd
    else:
        asd = ad
        bsd = bd

    if ret_idx:
        return asd, bsd, keep
    else:
        return asd, bsd


def valid(a, b):
    """Check whether `a` and `b` are not inf or nan"""
    return ~(np.isnan(a) | np.isinf(a) | np.isnan(b) | np.isinf(b))


def norm(a, ref1, ref2):
    """
    Normalize `a` with min/max values of `ref1`, using all elements of
    `ref1` where the `ref1` and `ref2` are not nan or inf"""
    ref = ref1[valid(ref1, ref2)]
    return (a-ref.min())/(ref.max()-ref.min())



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from ..util import hashfile  # noqa: F401

from .check import check_dataset  # noqa: F401
from .config import Configuration  # noqa: F401
from .core import RTDCBase  # noqa: F401
from .fmt_dict import RTDC_Dict  # noqa: F401
from .fmt_hdf5 import RTDC_HDF5  # noqa: F401
from .fmt_hierarchy import RTDC_Hierarchy  # noqa: F401
from .fmt_tdms import RTDC_TDMS  # noqa: F401
from .load import new_dataset  # noqa: F401
from .write_hdf5 import write  # noqa: F401



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import numpy as np

from dclab.features import emodulus


def test_simple_emod():
    x = np.linspace(0, 250, 100)
    y = np.linspace(0, 0.1, 100)
    x, y = np.meshgrid(x, y)

    emod = emodulus.get_emodulus(area_um=x,
                                 deform=y,
                                 medium="CellCarrier",
                                 channel_width=30,
                                 flow_rate=0.16,
                                 px_um=0,  # without pixelation correction
                                 temperature=23)

    assert np.allclose(emod[10, 50], 1.1875799054283109)
    assert np.allclose(emod[50, 50], 0.55281291845478731)
    assert np.allclose(emod[80, 50], 0.45678187644969814)

    assert np.allclose(emod[10, 80], 1.5744560306483262)
    assert np.allclose(emod[50, 80], 0.73534561544655519)
    assert np.allclose(emod[80, 80], 0.60737083178222251)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Test tdms file format"""
from __future__ import print_function

import os
import pathlib
import shutil
import sys
import tempfile

import numpy as np
import nptdms
import pytest

from dclab import new_dataset
import dclab.rtdc_dataset.fmt_tdms.naming

from helper_methods import retrieve_data, example_data_sets, cleanup


def test_compatibility_minimal():
    ds = new_dataset(retrieve_data("rtdc_data_minimal.zip"))
    assert ds.config["setup"]["channel width"] == 20
    assert ds.config["setup"]["chip region"].lower() == "channel"
    assert ds.config["setup"]["flow rate"] == 0.12
    assert ds.config["imaging"]["pixel size"] == 0.34
    cleanup()


@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="requires python3.6 or higher")
def test_compatibility_channel_width():
    # At some point, "Channel width" was repleaced by "Channel width [um]"
    path = retrieve_data("rtdc_data_minimal.zip")
    para = path.parent / "M1_para.ini"
    pardata = para.read_text()
    pardata = pardata.replace("Channel width = 20\n", "Channel width = 34\n")
    para.write_text(pardata)
    ds = new_dataset(path)
    assert ds.config["setup"]["channel width"] == 34
    cleanup()


def test_compatibility_shapein201():
    ds = new_dataset(retrieve_data("rtdc_data_shapein_v2.0.1.zip"))
    assert ds.config["setup"]["channel width"] == 20
    assert ds.config["setup"]["chip region"].lower() == "channel"
    assert ds.config["setup"]["software version"] == "ShapeIn 2.0.1"
    assert ds.config["imaging"]["pixel size"] == 0.34
    assert ds.config["imaging"]["flash duration"] == 2
    assert ds.config["experiment"]["date"] == "2017-10-12"
    assert ds.config["experiment"]["time"] == "12:54:31"
    cleanup()


def test_contains_non_scalar():
    ds1 = new_dataset(retrieve_data("rtdc_data_traces_video.zip"))
    assert "contour" in ds1
    assert "image" in ds1
    assert "mask" in ds1
    assert "trace" in ds1
    ds2 = new_dataset(retrieve_data("rtdc_data_minimal.zip"))
    assert "image" not in ds2
    assert "trace" not in ds2
    ds3 = new_dataset(retrieve_data("rtdc_data_shapein_v2.0.1.zip"))
    assert "contour" not in ds3
    assert "image" not in ds3
    assert "mask" not in ds3
    assert "trace" not in ds3
    cleanup()


def test_contour_basic():
    ds = new_dataset(retrieve_data(example_data_sets[1]))
    assert len(ds["contour"]) == 12
    assert np.allclose(np.average(ds["contour"][0]), 38.488764044943821)
    assert ds["contour"]._initialized
    cleanup()


def test_contour_corrupt():
    path = pathlib.Path(retrieve_data(example_data_sets[1]))
    cpath = path.with_name("M1_contours.txt")
    # remove initial contours
    with cpath.open("r") as fd:
        lines = fd.readlines()
    with cpath.open("w") as fd:
        fd.writelines(lines[183:])
    ds = new_dataset(path)
    try:
        ds["contour"].determine_offset()
    except dclab.rtdc_dataset.fmt_tdms.event_contour.ContourIndexingError:
        pass
    else:
        assert False
    cleanup()


def test_contour_naming():
    # Test that we always find the correct contour name
    ds = new_dataset(retrieve_data(example_data_sets[0]))
    dp = pathlib.Path(ds.path).resolve()
    dn = dp.parent
    contfile = dn / "M1_0.120000ul_s_contours.txt"
    contfileshort = dn / "M1_contours.txt"
    contfileexact = dn / "M1_2us_70A_0.120000ul_s_contours.txt"
    del ds

    # Test for perfect match
    # "M1_2us_70A_0.120000ul_s_contours.txt" should have priority over
    # "M1_contours.txt" and "M1_0.120000ul_s_contours.txt".
    shutil.copy(str(contfile), str(contfileshort))
    shutil.copy(str(contfile), str(contfileexact))
    ds2 = new_dataset(dp)
    assert str(ds2["contour"].identifier) == str(contfileexact)
    assert not np.allclose(ds2["contour"][1], 0)
    del ds2

    # Check if "M1_contours.txt" is used if the other is not
    # there.
    contfileshort.unlink()
    contfileexact.unlink()
    contfile.rename(contfileshort)
    ds3 = new_dataset(dp)
    assert str(ds3["contour"].identifier) == str(contfileshort)
    del ds3
    contfileshort.rename(contfile)

    # Create M10 file
    with (dn / "M10_contours.txt").open(mode="w"):
        pass
    ds4 = new_dataset(dp)
    assert str(ds4["contour"].identifier) == str(contfile)
    del ds4

    # Check when there is no contour file
    os.remove(str(contfile))
    # This will issue a warning that no contour data was found.
    ds5 = new_dataset(dp)
    assert ds5["contour"].identifier is None
    cleanup()


def test_contour_negative_offset():
    ds = new_dataset(retrieve_data(example_data_sets[1]))
    ds["contour"][0]
    ds["contour"].event_offset = 1
    assert np.all(ds["contour"][0] == np.zeros((2, 2), dtype=int))
    cleanup()


def test_contour_not_initialized():
    ds = new_dataset(retrieve_data(example_data_sets[1]))
    assert not ds["contour"]._initialized
    cleanup()


def test_fluorescence_config():
    ds1 = new_dataset(retrieve_data("rtdc_data_minimal.zip"))
    assert "fluorescence" not in ds1.config
    ds2 = new_dataset(retrieve_data("rtdc_data_traces_2flchan.zip"))
    assert "fluorescence" in ds2.config
    cleanup()


def test_image_basic():
    ds = new_dataset(retrieve_data(example_data_sets[1]))
    # Transition image
    assert np.allclose(ds["image"][0], 0)
    # Real image
    assert np.allclose(np.average(ds["image"][1]), 45.1490478515625)
    cleanup()


def test_image_column_length():
    ds = new_dataset(retrieve_data(example_data_sets[1]))
    assert len(ds["image"]) == 3
    cleanup()


def test_image_corrupt():
    path = pathlib.Path(retrieve_data(example_data_sets[1]))
    vpath = path.with_name("M1_imaq.avi")
    # create empty video file
    vpath.unlink()
    vpath.touch()
    try:
        new_dataset(path)
    except dclab.rtdc_dataset.fmt_tdms.event_image.InvalidVideoFileError:
        pass
    else:
        assert False
    cleanup()


def test_image_out_of_bounds():
    ds = new_dataset(retrieve_data(example_data_sets[1]))
    try:
        ds["image"][5]
    except IndexError:
        pass
    else:
        raise ValueError("IndexError should have been raised!")
    cleanup()


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'ancillaries.ancillary_feature.'
                            + 'BadFeatureSizeWarning')
def test_large_fov():
    ds = new_dataset(retrieve_data(example_data_sets[3]))
    # initial image is missing
    assert np.allclose(ds["image"][0], 0)
    # initial contour is empty
    assert np.allclose(ds["contour"][0], 0)
    # maximum of contour is larger than 255 (issue #167)
    assert ds["contour"][1].max() == 815
    # compute brightness with given contour
    # Remove the brightness column and let it recompute
    # using the ancillary columns. Besides testing the
    # correct positioning of the contour, this is a
    # sanity test for the brightness computation.
    bavg = ds._events.pop("bright_avg")
    bcom = ds["bright_avg"]
    assert np.allclose(bavg[1], bcom[1])
    cleanup()


def test_load_tdms_all():
    for ds in example_data_sets:
        tdms_path = retrieve_data(ds)
        ds = new_dataset(tdms_path)
    cleanup()


def test_load_tdms_avi_files():
    tdms_path = retrieve_data(example_data_sets[1])
    edest = pathlib.Path(tdms_path).parent
    ds1 = new_dataset(tdms_path)
    assert pathlib.Path(ds1["image"].video_file).name == "M1_imaq.avi"
    shutil.copyfile(str(edest / "M1_imaq.avi"),
                    str(edest / "M1_imag.avi"))
    ds2 = new_dataset(tdms_path)
    # prefer imag over imaq
    assert pathlib.Path(ds2["image"].video_file).name == "M1_imag.avi"
    shutil.copyfile(str(edest / "M1_imaq.avi"),
                    str(edest / "M1_test.avi"))
    ds3 = new_dataset(tdms_path)
    # ignore any other videos
    assert pathlib.Path(ds3["image"].video_file).name == "M1_imag.avi"
    os.remove(str(edest / "M1_imaq.avi"))
    os.remove(str(edest / "M1_imag.avi"))
    ds4 = new_dataset(tdms_path)
    # use available video if ima* not there
    assert pathlib.Path(ds4["image"].video_file).name == "M1_test.avi"
    cleanup()


def test_load_tdms_simple():
    tdms_path = retrieve_data(example_data_sets[0])
    ds = new_dataset(tdms_path)
    assert ds.filter.all.shape[0] == 156
    cleanup()


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'ancillaries.ancillary_feature.'
                            + 'BadFeatureSizeWarning')
def test_mask_basic():
    ds = new_dataset(retrieve_data(example_data_sets[1]))
    assert len(ds["mask"]) == 12
    # Test mask computation by averaging brightness and comparing to
    # the ancillary feature "bright_avg".
    bavg1 = ds["bright_avg"][1]
    bavg2 = np.mean(ds["image"][1][ds["mask"][1]])
    assert np.allclose(bavg1, bavg2), "mask is correctly computed from contour"
    cleanup()


def test_mask_img_shape_1():
    path = retrieve_data("rtdc_data_traces_video.zip")
    # shape from image data
    with new_dataset(path) as ds:
        assert ds["mask"]._img_shape == (96, 256)
    cleanup()


def test_mask_img_shape_2():
    path = retrieve_data("rtdc_data_traces_video.zip")
    path.with_name("M1_imaq.avi").unlink()
    with new_dataset(path) as ds:
        # shape from config ("roi size x", "roi size y")
        assert ds["mask"]._img_shape == (96, 256)
    # no shape available
    with new_dataset(path) as ds:
        ds.config["imaging"].pop("roi size x")
        ds.config["imaging"].pop("roi size y")
        assert ds["mask"]._img_shape == (0, 0)
        assert len(ds["mask"]) == 0
    cleanup()


def test_mask_img_wrong_config_shape_1():
    path = retrieve_data("rtdc_data_traces_video.zip")
    with new_dataset(path) as ds:
        # deliberately set wrong size in ROI (fmt_tdms tries image shape first)
        ds.config["imaging"]["roi size x"] = 200
        ds.config["imaging"]["roi size y"] = 200
        assert ds["mask"]._img_shape == (96, 256)
    cleanup()


def test_mask_img_wrong_config_shape_2():
    path = retrieve_data("rtdc_data_traces_video.zip")
    path.with_name("M1_imaq.avi").unlink()
    with new_dataset(path) as ds:
        # deliberately set wrong size in ROI (fmt_tdms tries image shape first)
        ds.config["imaging"]["roi size x"] = 200
        ds.config["imaging"]["roi size y"] = 200
        assert ds["mask"]._img_shape == (200, 200)
    cleanup()


def test_naming_valid():
    for key in dclab.rtdc_dataset.fmt_tdms.naming.dclab2tdms:
        assert key in dclab.definitions.feature_names


def test_parameters_txt():
    ds = new_dataset(retrieve_data("rtdc_data_frtdc_parameters.zip"))
    assert ds.config["setup"]["module composition"] == "Cell_Flow_2, Fluor"
    assert ds.config["setup"]["software version"] == "fRT-DC V0.80 150601"
    assert ds.config["setup"]["identifier"] == "47 red angels"
    assert ds.config["setup"]["medium"] == "other"
    assert ds.config["setup"]["flow rate sample"] == 0.01
    assert ds.config["setup"]["flow rate sheath"] == 0.03
    assert ds.config["imaging"]["pixel size"] == 0.34
    assert ds.config["imaging"]["flash duration"] == 2
    assert ds.config["fluorescence"]["samples per event"] == 1000
    assert ds.config["fluorescence"]["sample rate"] == 1000000
    assert ds.config["fluorescence"]["signal min"] == -1
    assert ds.config["fluorescence"]["signal max"] == 1
    assert ds.config["fluorescence"]["trace median"] == 21
    assert ds.config["fluorescence"]["laser 1 power"] == 5
    assert ds.config["fluorescence"]["laser 2 power"] == 0
    assert ds.config["fluorescence"]["laser 3 power"] == 0
    cleanup()


def test_pixel_size():
    path = retrieve_data("rtdc_data_minimal.zip")
    para = path.parent / "M1_para.ini"
    data = para.open("r").read()
    newdata = data.replace("Pix Size = 0.340000", "Pix Size = 0.120000")
    with para.open("w") as fd:
        fd.write(newdata)
    ds = new_dataset(path)
    assert ds.config["imaging"]["pixel size"] == 0.12
    cleanup()


def test_project_path():
    tfile = retrieve_data(example_data_sets[0])
    ds = dclab.new_dataset(tfile)
    assert ds.hash == "69733e31b005c145997fac8a22107ded"
    assert ds.format == "tdms"
    tpath = pathlib.Path(tfile).resolve()
    a = dclab.rtdc_dataset.fmt_tdms.get_project_name_from_path(str(tpath))
    b = dclab.rtdc_dataset.fmt_tdms.get_project_name_from_path(
        str(tpath.parent))
    assert a == b
    c = dclab.rtdc_dataset.fmt_tdms.get_project_name_from_path(
        str(tpath.parent / "online" / tpath.name))
    d = dclab.rtdc_dataset.fmt_tdms.get_project_name_from_path(
        str(tpath.parent / "online" / "data" / tpath.name))
    e = dclab.rtdc_dataset.fmt_tdms.get_project_name_from_path(
        str(tpath.parent / "online" / "data"))

    assert a == e
    assert a == c
    assert a == d
    cleanup()


def test_trace_basic():
    ds = new_dataset(retrieve_data(example_data_sets[1]))
    msg = "traces should not be loaded into memory before first access"
    assert ds["trace"].__repr__().count("<not loaded into memory>"), msg
    assert len(ds["trace"]) == 2
    assert np.allclose(np.average(
        ds["trace"]["fl1_median"][0]), 287.08999999999997)
    cleanup()


def test_trace_import_fail():
    # make sure undefined trace data does not raise an error
    tdms_path = retrieve_data(example_data_sets[1])
    dclab.definitions.FLUOR_TRACES.append("peter")
    dclab.rtdc_dataset.fmt_tdms.naming.tr_data_map["peter"] = [u'ukwn', u'ha']
    new_dataset(tdms_path)
    # clean up
    dclab.rtdc_dataset.fmt_tdms.naming.tr_data_map.pop("peter")
    dclab.definitions.FLUOR_TRACES.pop(-1)
    cleanup()


def test_trace_methods():
    ds = new_dataset(retrieve_data(example_data_sets[1]))
    for k in list(ds["trace"].keys()):
        assert k in dclab.definitions.FLUOR_TRACES
    for k in ds["trace"]:
        assert k in dclab.definitions.FLUOR_TRACES
    assert ds["trace"].__repr__().count("<loaded into memory>")
    cleanup()


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'fmt_tdms.event_trace.'
                            + 'MultipleSamplesPerEventFound')  # desired
def test_trace_wrong_samples_per_event():
    """The "samples per event" should be a constant for trace data

    If this condition is not given, the trace data is considered
    unusable, but the rest of the data still makes sense
    (Philipp said that something must have gone wrong during writing
    of the trace data).
    """
    tdms = retrieve_data("rtdc_data_traces_video.zip")
    mdata = nptdms.TdmsFile(str(tdms))
    objects = [obj for obj in mdata.objects.values()]

    # blank write same data to test that modification works
    with nptdms.TdmsWriter(str(tdms)) as tdms_writer:
        tdms_writer.write_segment(objects)

    with dclab.new_dataset(tdms) as ds:
        assert "trace" in ds

    # modify objects
    sampleids = mdata.object("Cell Track", "FL1index").data
    sampleids[0] = 10
    sampleids[1] = 20
    sampleids[2] = 40
    objects2 = []
    for obj in objects:
        if obj.channel == "FL1index":
            nptdms.ChannelObject("Cell Track", "FL1index", sampleids)
        objects2.append(obj)

    with nptdms.TdmsWriter(str(tdms)) as tdms_writer:
        tdms_writer.write_segment(objects2)

    with dclab.new_dataset(tdms) as ds:
        assert "trace" not in ds
    cleanup()


def test_unicode_paths():
    path = retrieve_data(example_data_sets[1])
    path = pathlib.Path(path)
    pp = path.parent
    # create a unicode name
    pp2 = pathlib.Path(tempfile.mktemp(prefix="dclàb_tést_asgård_únícodè"))
    pp.rename(pp2)
    ds = new_dataset(pp2 / path.name)
    ds.__repr__()
    shutil.rmtree(str(pp2), ignore_errors=True)
    cleanup()


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import division, print_function, unicode_literals

from ... import features
from .ancillary_feature import AncillaryFeature


def compute_emodulus_legacy(mm):
    """This is how it was done in Shape-Out 1"""
    calccfg = mm.config["calculation"]
    model = calccfg["emodulus model"]
    assert model == "elastic sphere"
    medium = calccfg["emodulus medium"]
    viscosity = calccfg["emodulus viscosity"]
    temperature = mm.config["calculation"]["emodulus temperature"]
    if medium.lower() == "other":
        medium = viscosity
        temperature = None
    # compute elastic modulus
    emod = features.emodulus.get_emodulus(
        area_um=mm["area_um"],
        deform=mm["deform"],
        medium=medium,
        channel_width=mm.config["setup"]["channel width"],
        flow_rate=mm.config["setup"]["flow rate"],
        px_um=mm.config["imaging"]["pixel size"],
        temperature=temperature)
    return emod


def compute_emodulus_known_media(mm):
    """Only use known media and one temperature for all

    This is a special case in :func:`compute_emodulus_legacy`.
    """
    calccfg = mm.config["calculation"]
    model = calccfg["emodulus model"]
    assert model == "elastic sphere"
    medium = calccfg["emodulus medium"]
    if medium not in features.emodulus_viscosity.KNOWN_MEDIA:
        raise ValueError("Only the following media are supported: {}".format(
                         features.emodulus_viscosity.KNOWN_MEDIA))
    # compute elastic modulus
    emod = features.emodulus.get_emodulus(
        area_um=mm["area_um"],
        deform=mm["deform"],
        medium=medium,
        channel_width=mm.config["setup"]["channel width"],
        flow_rate=mm.config["setup"]["flow rate"],
        px_um=mm.config["imaging"]["pixel size"],
        temperature=mm.config["calculation"]["emodulus temperature"])
    return emod


def compute_emodulus_temp_feat(mm):
    """Use the "temperature" feature"""
    calccfg = mm.config["calculation"]
    model = calccfg["emodulus model"]
    assert model == "elastic sphere"
    medium = calccfg["emodulus medium"]
    assert medium != "other"
    # compute elastic modulus
    emod = features.emodulus.get_emodulus(
        area_um=mm["area_um"],
        deform=mm["deform"],
        medium=medium,
        channel_width=mm.config["setup"]["channel width"],
        flow_rate=mm.config["setup"]["flow rate"],
        px_um=mm.config["imaging"]["pixel size"],
        temperature=mm["temp"])
    return emod


def compute_emodulus_visc_only(mm):
    """The user entered the viscosity directly

    This is a special case in :func:`compute_emodulus_legacy`.
    """
    calccfg = mm.config["calculation"]
    model = calccfg["emodulus model"]
    assert model == "elastic sphere"
    viscosity = calccfg["emodulus viscosity"]
    # compute elastic modulus
    emod = features.emodulus.get_emodulus(
        area_um=mm["area_um"],
        deform=mm["deform"],
        medium=viscosity,
        channel_width=mm.config["setup"]["channel width"],
        flow_rate=mm.config["setup"]["flow rate"],
        px_um=mm.config["imaging"]["pixel size"],
        temperature=None)
    return emod


def register():
    # Please note that registering these things is a delicate business,
    # because the priority has to be chosen carefully.
    AncillaryFeature(feature_name="emodulus",
                     method=compute_emodulus_legacy,
                     req_features=["area_um", "deform"],
                     req_config=[["calculation", ["emodulus medium",
                                                  "emodulus model",
                                                  "emodulus temperature",
                                                  "emodulus viscosity"]],
                                 ["imaging", ["pixel size"]],
                                 ["setup", ["flow rate", "channel width"]]
                                 ],
                     priority=3)
    AncillaryFeature(feature_name="emodulus",
                     method=compute_emodulus_known_media,
                     req_features=["area_um", "deform"],
                     req_config=[["calculation", ["emodulus medium",
                                                  "emodulus model",
                                                  "emodulus temperature"]],
                                 ["imaging", ["pixel size"]],
                                 ["setup", ["flow rate", "channel width"]]
                                 ],
                     priority=2)
    AncillaryFeature(feature_name="emodulus",
                     method=compute_emodulus_visc_only,
                     req_features=["area_um", "deform"],
                     req_config=[["calculation", ["emodulus model",
                                                  "emodulus viscosity"]],
                                 ["imaging", ["pixel size"]],
                                 ["setup", ["flow rate", "channel width"]]
                                 ],
                     priority=1)
    AncillaryFeature(feature_name="emodulus",
                     method=compute_emodulus_temp_feat,
                     req_features=["area_um", "deform", "temp"],
                     req_config=[["calculation", ["emodulus medium",
                                                  "emodulus model"]],
                                 ["imaging", ["pixel size"]],
                                 ["setup", ["flow rate", "channel width"]]
                                 ],
                     priority=0)
